# -*- coding: utf-8 -*-
# flake8: noqa

"""
OpenGSQ Python Library
~~~~~~~~~~~~~~~~~~~

Python library for querying game servers

:copyright: (c) 2021 BattlefieldDuck
:license: MIT, see LICENSE for more details.

"""

from opengsq.protocols import *
